/******************************************************************************

	analyzer.c

	GPIB analyzer ("sniffer"). Used to trace communication between any other
	devices connected to the same GPIB bus. Requires a NEC7210 based IEEE488
	interface. Terminates either when being addressed over GPIB (talker or
	listener) or when the maximum number of 10000 samples is reached.
	
	(c) 2007 by Ansgar Kueckes (kueckes@alice-dsl.net)

******************************************************************************/

#ifdef __WIN32__
#include <windows.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>

#ifdef USE_DLPORTIO
#include "..\dlportio\dlportio.h"
#endif

/* 7210 read registers */
#define DIR  0x2E1
#define ISR1 0x6E1
#define ISR2 0xAE1
#define SPSR 0xEE1
#define ADSR 0x12E1
#define CPTR 0x16E1
#define ADR0 0x1AE1
#define ADR1 0x1EE1

/* 7210 write registers */
#define CDOR  0x2E1
#define IMR1  0x6E1
#define IMR2  0xAE1
#define SPMR  0xEE1
#define ADMR  0x12E1
#define AUXMR 0x16E1
#define ADR   0x1AE1
#define EOSR  0x1EE1

#ifndef __WIN32__
#define BYTE unsigned char
#define WORD unsigned short
#define DWORD unsigned long
#endif

void outp(unsigned short IO_port, unsigned char Value)
{
   asm( "mov %0, %%dx;"
        "mov %1, %%al;"
   	"out %%al, %%dx;"
   : //no output data
   :"r"(IO_port),"r"(Value)
   :"dx","al"
   );
}

unsigned char inp(unsigned short IO_port)
{ 
   BYTE data;
   asm( "mov %1, %%dx;"
   	"in %%dx, %%al;"
   	"mov %%al, %0;"
   :"=r"(data)
   :"r"(IO_port)
   :"dx","al"
   );
   return data;
}

#ifdef USE_DLPORTIO
  #ifdef __WIN32__
     #define OUTP(port, value)       DlPortWritePortUchar(port, value)
     #define INP(port)               DlPortReadPortUchar(port)
  #endif
#else
  #define OUTP(port, value)			outp(port, value)
  #define INP(port)		        	inp(port) 
#endif

void error()
{
   fprintf(stderr,"Error encountered.");
   exit(0);
}

void output_binary(BYTE value)
{
   unsigned long mask = 0x80;
   int i;

   for (i=0; i<8; i++) {
      printf ((value & mask) ? "1" : "0");
      mask >>= 1;
   }
}

void gpib_initialize()
{
   OUTP(AUXMR, 0x2);    /* chip reset */
   OUTP(IMR1, 0x0);     /* clear both mask registers */
   OUTP(IMR2, 0x0);
   OUTP(ADMR, 0x0);     /* clear mode register */
   OUTP(AUXMR, 0x0);    /* release from idle state */
}

#define MAX_SIZE 10000

int main()
{
   BYTE adsr[MAX_SIZE];
   BYTE cptr[MAX_SIZE];
   DWORD tick[MAX_SIZE];
   DWORD count;
   BYTE tmp_data, tmp_status;
   int atn, tmp_atn;
   int i,j;
   unsigned char a;
   clock_t clk;

   printf("initializing\n");
   gpib_initialize();

   /* disable system controller */
   OUTP(AUXMR, 0x14);

   /* set primary address 20 */
   OUTP(ADR, 0x14);
   OUTP(ADR, 0x94);

   /* enable talker/listener */
   OUTP(ADMR, 0x31);

   /* clear pon */
   OUTP(AUXMR, 0x0);

   tmp_status = 0;
   tmp_data = 0;
   count = 0;
   i=0;
   clk = clock();

   /* read in samples */
   while (i<MAX_SIZE) {

     cptr[i]=INP(CPTR);
     adsr[i]=INP(ADSR);
     count++;

     /* record changes only */
     if ((tmp_status != adsr[i]) || (tmp_data != cptr[i])) {
	tmp_status = adsr[i];
	tmp_data = cptr[i];
        tick[i]=count;
	/* check for being addressed to talk or listen (end condition) */
	if (tmp_status & 0xc) goto print;
	i++;
     }
   }

   print:
   /* printout statistics */
   clk = clock()-clk;
   printf("total recording time: %.2f sec (%lu samples)\n", (double)(clk)/CLOCKS_PER_SEC, count);
   printf("one sample is aprox. %.2f us (%.0f ksamples/s)\n\n",
   	(double)(clk)*1000000/CLOCKS_PER_SEC/count,
   	(double)(count)/1000/clk*CLOCKS_PER_SEC);

   /* printout samples */
   tmp_atn = 0;
   for (j=0; j<i; j++) {
	/* test for new command */
	atn = !(adsr[j] & 0x40);
	if (tmp_atn != atn) printf("\n");
	tmp_atn = atn;
        a = cptr[j] & 0x7f;
        if ((a<32) || (a>127)) a = ' ';
	printf ("%5d: ",j);
	output_binary(adsr[j]);
	printf(" ");
	output_binary(cptr[j]);
	printf(" ($%02X,$%02X) '%c'", adsr[j], cptr[j], a);
	if (j>0) printf(" [%lu]", tick[j]-tick[j-1]);
	else printf(" [%lu]", tick[j]);
	if (atn) {
	   if (cptr[j]==0x01) printf("\tGTL\n");
	   else if (cptr[j]==0x04) printf("\tSDC\n");
	   else if (cptr[j]==0x05) printf("\tPPC\n");
	   else if (cptr[j]==0x08) printf("\tGET\n");
	   else if (cptr[j]==0x09) printf("\tTCT\n");
	   else if (cptr[j]==0x11) printf("\tLLO\n");
	   else if (cptr[j]==0x14) printf("\tDCL\n");
	   else if (cptr[j]==0x15) printf("\tPPU\n");
	   else if (cptr[j]==0x18) printf("\tSPE\n");
	   else if (cptr[j]==0x19) printf("\tSPD\n");
	   else if (cptr[j]==0xbf) printf("\tUNL\n");
	   else if (cptr[j]==0xdf) printf("\tUNT\n");
	   else if ((cptr[j]&0x60)==0x20) printf("\tLAD %d\n", cptr[j]&0x1f);
	   else if ((cptr[j]&0x60)==0x40) printf("\tTAD %d\n", cptr[j]&0x1f);
	   else if ((cptr[j]&0x60)==0x60) printf("\tSEC $%02X\n", cptr[j]&0x1f);
	   else printf("\n");
	}
	else printf ("\t$%02X\n", cptr[j]);
   }
   return 1;
}
