/******************************************************************************

	bmpconvert.c

	Converts between generic 9845 graphics image formats and the Windows
	BMP format.
	
	All image files must have dimensions 560 columns and 455 rows.
	
	BMP formats are supported with 1, 4, 8 and 24 bits per pixel, however
	it is recommended to use 1 bit per pixel for bilevel and 4 bits per
	pixel for color images.
	
	Assumed color map for indexed files is BLK, RED, GRN, YEL, BLU, MAG, CYA,
	WHT (in that order). All other color maps will produce false colors.
	Due to the limited color space of the 9845, it is recommended to apply an
	adaquate dithering scheme before transferring photographs into a 9845
	format. Again note that you *must* use the above color palette in order
	to produce images with correctly indexed colors.
	
	bmpconvert just does the necessary conversions, it does not adjust
	the size of the image nor does it check for file format integrity. Use your
	favourite image editing software to resize and/or dither the source image
	bitmap.
	
	Both DATA and BDAT file formats are supported, however it is recommended
	to use BDAT format since it consumes half the space and one 10th of the
	time for storing & retrieving images compared to the DATA format (full
	color saves or loads need just about 20 seconds in BDAT format).
	
	HP color images have to be split into multiple files due to limitations of
	array size in HP BASIC. We assume the color image data to be provided
	in three files of ca. 16K each, with a maximum file name length of 5
	plus "1" for the first, "2" for the second and "3" for the last file.
	
	Vice versa, if color images are produced, they will be spread over three
	files accordingly.
	
	(c) 2008 by Ansgar Kueckes (www.hp9845.net)

******************************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <windows.h>

#pragma pack(1)

typedef struct {
	WORD id;
	WORD data;
} HPINT;

typedef struct {
	BYTE peBlue;
	BYTE peGreen;
	BYTE peRed;
	BYTE peFlags;	
} Palette_Entry;

#pragma pack()

typedef enum { DATA, BDAT } HPFormat;
typedef enum { STANDARD, ENHANCED, COLOR } HPSystem;

void print_usage(char *command)
{
	printf("Usage: %s [<options>] <infile> <outfile>\n", command);
	printf("       <infile>   input image file\n");
	printf("       <outfile>  output image file\n");
	printf("\nwith <options>:\n");
	printf("       -bdat      produce BDAT format (default)\n");
	printf("       -data      produce DATA format\n");
	printf("       -standard  produce monochrome format for 98750A displays (default)\n");
	printf("       -enhanced  produce monochrome format for 98780A displays\n");
	printf("       -color     produce color data (needs two separate 9845C files)\n");
	printf("       -info      output file info only\n");
	return;
}

/*
	append "2" to filename and truncate as necessary
*/
void create_filename(char *new_filename, char *filename, int count)
{
	char *extension;
	int base;
	
	/* append "count" to filename, truncate filename if necessary */
	new_filename[0] = 0;
	extension = strchr(filename, '.');
	if (extension) {
		base = extension - filename;
		if (base > 5) base = 5;
		else if (filename[base-1] == '1') base--;
		strncpy(new_filename, filename, base);
		new_filename[base] = 48 + count;
		new_filename[base+1] = 0;
		strcat(new_filename, extension);
	}
	else {
		base = strlen(filename);
		if (base > 5) base = 5;
		else if (filename[base-1] == '1') base--;
		strncpy(new_filename, filename, base);
		new_filename[base] = 48 + count;
		new_filename[base+1] = 0;
	}
}

/*
	convert block of 16 pixels from BMP to HP 9845 format
*/
void convert_pixel_bmp2hp(BYTE *buffer, HPINT *ret_plane, int bmp_depth, int hp_depth)
{
	int i,j,pos,offset;
	WORD pixel_mask, color_mask;

	/* initialize plane data */	
	for (i=0; i<hp_depth; i++) {
		ret_plane[i].id = 0x0a00;
		ret_plane[i].data = 0;
	}
	
	/* process pixels */
	for (i=0; i<16; i++) {
		if (i<8) pixel_mask = 0x80 >> i;
		else  pixel_mask = 0x8000 >> (i-8);
		switch (bmp_depth) {
		case 1: /* bilevel: maps to green (second plane) */
			pos = i / 8;
			offset = i % 8;
			color_mask = 0x80 >> offset;
			if (buffer[pos] & color_mask) {
				if (hp_depth == 1) ret_plane[0].data |= pixel_mask;
				else ret_plane[1].data |= pixel_mask;
			}
			break;
		case 4: /* indexed */
		case 8:
			pos = (i * bmp_depth) / 8;
			offset = ((15 - i) * bmp_depth) % 8;
			if (hp_depth == 1) {
				// any value other than 0 will set the pixel
				if (bmp_depth == 4) color_mask = 0xf << offset;
				else color_mask = 0xff;
				if (buffer[pos] & color_mask) ret_plane[0].data |= pixel_mask;
			}
			else {
				// assume first 8 values in color map as BLK, RED, GRN,
				// YEL, BLU, MAG, CYA, WHT (all other bits ignored)
				for (j=0; j<hp_depth; j++) {
					color_mask = 0x1 << (j + offset);
					if (buffer[pos] & color_mask) ret_plane[j].data |= pixel_mask;
				}
			}
			break;
		case 24: /* true color: use color value of 128 as threshold */
			for (j=0; j<hp_depth; j++) {
				if (buffer[(i*3)+(2-j)] > 127) ret_plane[j].data |= pixel_mask;
			}
			break;
		}
	}
}

/*
	convert block of 16 pixels from HP 9845 to BMP format
*/
void convert_pixel_hp2bmp(HPINT *plane, BYTE *ret_buffer, int hp_depth, int bmp_depth)
{
	int i,j,pos,offset;
	WORD pixel_mask, color_mask;

	/* initialize buffer data */	
	for (i=0; i<(16*bmp_depth)/8; i++) {
		ret_buffer[i] = 0;
	}
	
	/* process pixels */
	for (i=0; i<16; i++) {
		if (i<8) pixel_mask = 0x80 >> i;
		else  pixel_mask = 0x8000 >> (i-8);
		switch (bmp_depth) {
		case 1: /* bilevel: any value will map */
			pos = i / 8;
			offset = i % 8;
			color_mask = 0x1 << offset;
			for (j=0; j<hp_depth; j++) {
				if (plane[j].data & pixel_mask) ret_buffer[pos] |= color_mask;
			}
			break;
		case 4: /* indexed */
		case 8:
			pos = (i * bmp_depth) / 8;
			offset = ((15 - i) * bmp_depth) % 8;
			if (hp_depth == 1) {
				// set pixel map to last entry in color map
				if (bmp_depth == 4) color_mask = 0xf << offset;
				else color_mask = 0xff;
				if (plane[0].data & pixel_mask) ret_buffer[pos] |= color_mask;
			}
			else {
				// assume first 8 values in color map as BLK, RED, GRN, YEL,
				//  BLU, MAG, CYA, WHT (all other bits left zero)
				for (j=0; j<hp_depth; j++) {
					color_mask = 0x1 << (j + offset);
					if (plane[j].data & pixel_mask) ret_buffer[pos] |= color_mask;
				}
			}
			break;
		case 24: /* true color: set color value to 255 */
			for (j=0; j<hp_depth; j++) {
				if (plane[j].data & pixel_mask) ret_buffer[(i*3)+(2-j)] = 255;
			}
			break;
		}
	}	
}

/*
	pad current output stream to defined record size
*/
void pad_record(FILE *ofp, int recordsize, BYTE value)
{
	int i,count;
	fpos_t pos;
	
	fgetpos(ofp, &pos);
	count = pos % recordsize;
	if (count == 0) return;
	count = recordsize - count;
	for (i=0; i<count; i++) fputc(value, ofp);
}

/*
	convert Windows BMP graphics file to HP 9845 graphics format (output will be
	split in two separate files in case HP 9845 color format is requested)
*/
void bmp2hp(HPFormat target_format, HPSystem target_system, FILE *ifp, FILE **ofp, DWORD offset)
{
  BITMAPINFOHEADER infoheader;
  BITMAPCOREHEADER coreheader;
  DWORD header_size;
  BYTE buffer[256];
  HPINT plane[3];
  BYTE bdat_header[256];
  int part,current,num_rows,rowsize,height,depth,x,y,i;
  
  fseek(ifp, 14, SEEK_SET);
  fread(&header_size, sizeof(header_size), 1, ifp);
  fseek(ifp, 14, SEEK_SET);

  /* perform the necessary checks */
  switch(header_size) {
  case 40:	fread(&infoheader, sizeof(infoheader), 1, ifp);
			if ((infoheader.biWidth != 560) || (abs(infoheader.biHeight) != 455)) {
				fprintf(stderr, "Error: input image dimensions do not match (must be 560x455)\n");
				exit(-1);
			}
			if (infoheader.biCompression!= BI_RGB) {
				fprintf(stderr, "Error: input image uses compression (must be uncompressed)\n");
				exit(-1);
			}
			depth = infoheader.biBitCount;
			height = infoheader.biHeight;
			break;
  case 12:	fread(&coreheader, sizeof(coreheader), 1, ifp);
			if ((coreheader.bcWidth != 560) || (abs(coreheader.bcHeight) != 455)) {
				fprintf(stderr, "Error: input image dimensions do not match (must be 560x455)\n");
				exit(-1);
			}
			depth = coreheader.bcBitCount;
			height = coreheader.bcHeight;
 			break;
  default:	fprintf(stderr, "Warning: unsupported BMP version\n");
  }

  if ((depth != 1) && (depth != 4) && (depth != 8) && (depth != 24)) {
	fprintf(stderr, "Error: unsupported color depth in BMP file (must be 1, 4, 8 or 24 bit/pixel)\n");
	exit(-1);
  }
  
  if (depth == 1) rowsize = 72;
  else rowsize = 560 * depth / 8;
  	
  if (target_format == BDAT) {
	/* prepare header */
	memset(bdat_header, 0, 256);
	bdat_header[0] = 0x00;
	bdat_header[1] = 0x04;
	bdat_header[2] = 0xff;
	bdat_header[3] = 0xff;
	bdat_header[4] = 0x72;
	bdat_header[5] = 0x01;
	bdat_header[8] = 0xff;
	bdat_header[9] = 0xff;
  }

  /* now start the conversion process */
  switch(target_system) {
  case COLOR:

	current = 0;
	
	if (target_format == BDAT) printf("BMP->BDAT\n");
	else printf("BMP->DATA\n");

	/* write all three parts */
	for (part=0; part<3; part++) {

		/* write header */
		if (target_format == BDAT) {
			switch(part) {
			case 0:
				bdat_header[6] = 0x3d;
				bdat_header[7] = 0xef;
				bdat_header[10] = 0x3d;
				bdat_header[11] = 0xef;
				bdat_header[12] = 0x3d;
				bdat_header[13] = 0xef;
				break;
			case 1:
			case 2:
				bdat_header[6] = 0x3e;
				bdat_header[7] = 0x58;
				bdat_header[10] = 0x3e;
				bdat_header[11] = 0x58;
				bdat_header[12] = 0x3e;
				bdat_header[13] = 0x58;
				break;
			}
			fwrite(bdat_header, 1, 256, ofp[part]);
		}
			
		/* convert */
		if (part == 0) num_rows = 151;
		else num_rows = 152;
		for (i=0; i<num_rows; i++) {
			if (height < 0) fseek(ifp, offset+(current*rowsize), SEEK_SET);
			else fseek(ifp, offset+((454-current)*rowsize), SEEK_SET);
			
			for (x=0; x<35; x++) {
				/* read in 16 pixel */
			  	fread(buffer, (16*depth)/8, 1, ifp);
				convert_pixel_bmp2hp(buffer, plane, depth, 3);
				switch (target_format) {
				case BDAT:
			  		fwrite(&plane[0].data, 2, 1, ofp[part]);
			  		fwrite(&plane[1].data, 2, 1, ofp[part]);
			  		fwrite(&plane[2].data, 2, 1, ofp[part]);
			  		break;
				case DATA:
			  		fwrite(plane, 12, 1, ofp[part]);
			  		break;
			  	}
			}
			current++;
		}
	
		/* write trailer */
		if (target_format == DATA) {
			/* write EOR */
			fputc(0x00, ofp[part]);
			fputc(0x1e, ofp[part]);
	
			/* write remaining stuff (probably not important) */
			fputc(0, ofp[part]);
			fputc(0, ofp[part]);
			plane[0].id = 0x0a00;
			plane[0].data = 0x0;
			fwrite(&plane[0], 4, 1, ofp[part]);
			fwrite(&plane[0], 4, 1, ofp[part]);
		}
	
		/* pad to 256-byte record border */
		pad_record(ofp[part], 256, 0);
	}
		
	break;
  
  case STANDARD:	/* bilevel */
  case ENHANCED:

	/* write header */		
	if (target_format == BDAT) {
		printf("BMP->BDAT\n");
		if (target_system == STANDARD) {
			bdat_header[6] = 0x3f;
			bdat_header[7] = 0xfd;
			bdat_header[10] = 0x3f;
			bdat_header[11] = 0xfd;
			bdat_header[12] = 0x3f;
			bdat_header[13] = 0xfd;
		} else {
			bdat_header[6] = 0x3e;
			bdat_header[7] = 0x35;
			bdat_header[10] = 0x3e;
			bdat_header[11] = 0x35;
			bdat_header[12] = 0x3e;
			bdat_header[13] = 0x35;
		}
		fwrite(bdat_header, 1, 256, ofp[0]);
		
		if (target_system == STANDARD) {
			/* write offset 0 (complemented = -1) */
			plane[0].data = 0xffff;
			fwrite(&plane[0].data, 2, 1, ofp[0]);
		}
	}
	else {
		printf("BMP->DATA\n");
		if (target_system == STANDARD) {
			/* write offset 0 (complemented = -1) */
			plane[0].id = 0x0a00;
			plane[0].data = 0xffff;
			fwrite(&plane[0], 4, 1, ofp[0]);
		}
	}
		
	/* convert image data */
	for (y=0; y<455; y++) {
		if (height < 0) fseek(ifp, offset+(y*rowsize), SEEK_SET);
		else fseek(ifp, offset+((454-y)*rowsize), SEEK_SET);
		
		for (x=0; x<35; x++) {
		  	fread(buffer, (16*depth)/8, 1, ifp);
			convert_pixel_bmp2hp(buffer, plane, depth, 1);
			switch(target_format) {
			case BDAT:
		  		fwrite(&plane[0].data, 2, 1, ofp[0]);
		  		break;
		  	case DATA:
		  		fwrite(&plane[0], 4, 1, ofp[0]);
		  		break;
		  	}
		}

		if (target_system == STANDARD) {
			/* write hidden data at end-of-row */
			plane[0].data = 0x0;
			if (target_format == BDAT)
				fwrite(&plane[0].data, 2, 1, ofp[0]);
			else
				fwrite(&plane[0], 4, 1, ofp[0]);
		}
	}

	/* write trailer */	
	if (target_format == DATA) {
		/* write EOR */
		fputc(0x00, ofp[0]);
		fputc(0x1e, ofp[0]);

		/* write remaining stuff */
		fputc(0, ofp[0]);
		fputc(0, ofp[0]);
		plane[0].id = 0x0a00;
		plane[0].data = 0x0;
		fwrite(&plane[0], 4, 1, ofp[0]);
		fwrite(&plane[0], 4, 1, ofp[0]);
	}

	/* pad to 256-byte record border */
	pad_record(ofp[0], 256, 0);

	break;
  }
}

/*
	convert HP 9845 graphics file to Windows BMP graphics format (input is
	taken from two separate files if HP 9845 color format is requested)
*/
void hp2bmp(HPFormat source_format, HPSystem source_system, FILE **ifp, FILE *ofp)
{
  BITMAPFILEHEADER fileheader;
  BITMAPINFOHEADER infoheader;
  Palette_Entry color_palette[16];
  BYTE bdat_header[256];
  WORD frame_buffer[47775];
  HPINT plane[3];
  int x,y,i,index,part,current,num_rows;
  int table_size, first_chunk, second_chunk;
  BYTE buffer[256];
  
  /* prepare common header data (54 bytes total) */
  fileheader.bfType = 0x4d42;
  fileheader.bfReserved1 = 0;
  fileheader.bfReserved2 = 0;

  infoheader.biSize = 40;
  infoheader.biWidth = 560;
  infoheader.biHeight = 455;
  infoheader.biPlanes = 1;
  infoheader.biCompression = 0;
  infoheader.biXPelsPerMeter = 0;
  infoheader.biYPelsPerMeter = 0;
  infoheader.biClrUsed = 0;
  infoheader.biClrImportant = 0;
  
  /* clear color palette */
  memset(color_palette, 0, sizeof(color_palette));

  switch (source_system) {
  case COLOR:
	/* update BMP header */
	fileheader.bfSize = 127518;
	fileheader.bfOffBits = 118;
	infoheader.biBitCount = 4;
	infoheader.biSizeImage = 127400;
		
	for (i=0; i<16; i++) {
		color_palette[i].peRed = (i & 0x1) ? 255 : 0;
		color_palette[i].peGreen = (i & 0x2) ? 255 : 0;
		color_palette[i].peBlue = (i & 0x4) ? 255 : 0;
	}

	/* write headers & color palette */
	fwrite(&fileheader, sizeof(fileheader), 1, ofp);
	fwrite(&infoheader, sizeof(infoheader), 1, ofp);
	fwrite(color_palette, sizeof(PALETTEENTRY), 16, ofp);

	/* now start the conversion process */
	if (source_format == BDAT) {
	  	printf("BDAT->BMP\n");
	  	
		index = 0;
		memset(frame_buffer, 0, sizeof(frame_buffer));

	  	/* process all three files */
	  	for (part=0; part<3; part++) {
	  	
		  	/* check BDAT header */
			fseek(ifp[part], 0, SEEK_SET);
			fread(bdat_header, 256, 1, ifp[part]);
			table_size = 256 * bdat_header[10] + bdat_header[11];
			first_chunk = 256 * bdat_header[12] + bdat_header[13];
			second_chunk = 256 * bdat_header[14] + bdat_header[15];
			
			/* limit data to dimensioned array size */
			if (table_size < first_chunk) first_chunk = table_size;
			if (table_size < (first_chunk + second_chunk)) second_chunk = table_size - first_chunk;
	
			/* read in chunks */
			for (i=0; i<first_chunk; i++) {
				fread(&frame_buffer[index++], 2, 1, ifp[part]);
			}
	
			if (second_chunk) {
				fseek(ifp[part], (((first_chunk * 2) + 511) / 256) * 256, SEEK_SET);
				for (i=0; i<second_chunk; i++) fread(&frame_buffer[index++], 2, 1, ifp[part]);
			}			
		}

		if (index != 47775) {
			printf("Warning: input data doesn't match (%d instead of 47775 words)\n", index);
		}	

		/* now do the conversion */
		for (y=454; y>=0; y--) {
			int pos;
		
			/* read graphics row */
			for (x=0; x<35; x++) {
				pos = (y * 105) + (x * 3);
				plane[0].data = frame_buffer[pos];
				plane[1].data = frame_buffer[pos + 1];
				plane[2].data = frame_buffer[pos + 2];
				convert_pixel_hp2bmp(plane, buffer, 3, 4);
				fwrite(buffer, 8, 1, ofp);
			}
		}
	}
	else { /* DATA format */
		printf("DATA->BMP\n");

		for (part=2; part>=0; part--) {

			if (part == 0) num_rows = 151;
			else num_rows = 152;

		  	/* second part first (reverse order) */
			for (i=num_rows-1; i>=0; i--) {
			
				fseek(ifp[part], i*420, SEEK_SET);
				
				/* read graphics row */
				for (x=0; x<35; x++) {
					fread(plane, 12, 1, ifp[part]);
					convert_pixel_hp2bmp(plane, buffer, 3, 4);
					fwrite(buffer, 8, 1, ofp);
				}
				current--;
			}
		}
	}
	break;
  
  case STANDARD:	/* bilevel */
  case ENHANCED:
  	  /* update BMP header */
	  fileheader.bfSize = 32822;
	  fileheader.bfOffBits = 62;
	  infoheader.biBitCount = 1;
	  infoheader.biSizeImage = 32760;

	  /* pixel color green */
	  color_palette[1].peGreen = 255;

	  /* write headers & color palette */
	  fwrite(&fileheader, sizeof(fileheader), 1, ofp);
	  fwrite(&infoheader, sizeof(infoheader), 1, ofp);
	  fwrite(color_palette, sizeof(PALETTEENTRY), 2, ofp);

	  /* now start the conversion process */
	  if (source_format == BDAT) {
	  	printf("BDAT->BMP\n");
		for (y=454; y>=0; y--) {
		
			if (source_system == STANDARD) fseek(ifp[0], 258+(y*72), SEEK_SET);
			else fseek(ifp[0], 256+(y*70), SEEK_SET);
			
			/* read graphics line*/
			for (x=0; x<35; x++) {
				fread(&plane[0].data, 2, 1, ifp[0]);
				fwrite(&plane[0].data, 2, 1, ofp);
			}
			
			/* pad to 4 byte border */
			fputc(0, ofp);
			fputc(0, ofp);
		}
	  }
	  else {
	  	printf("DATA->BMP\n");
		for (y=454; y>=0; y--) {
		
			if (source_system == STANDARD) fseek(ifp[0], 4+(y*144), SEEK_SET);
			else fseek(ifp[0], y*140, SEEK_SET);
			
			/* read graphics line*/
			for (x=0; x<35; x++) {
				fread(&plane[0], 4, 1, ifp[0]);
				fwrite(&plane[0].data, 2, 1, ofp);
			}
			
			/* pad to 4 byte border */
			fputc(0, ofp);
			fputc(0, ofp);
		}
	  }
      break;
  }
}

int main(int argc, char *argv[])
{
  int i,count;
  BOOL info = FALSE;
  FILE *ifp[3];
  FILE *ofp[3];
  char *infile = 0;
  char *outfile = 0;
  char tmp[MAX_PATH];
  HPFormat hp_format = BDAT;
  HPSystem hp_system = STANDARD;
  HPINT hpint;
  BYTE buffer[256];
  DWORD offset;
  WORD id;
  int tablesize;
  
  if (argc < 2) { print_usage(argv[0]); exit(0); }

  /* command line processing */
  for ( count = 1; count < argc ; count++)
  {
	if (!strcmp(argv[count], "-bdat"))
		 { hp_format = BDAT; continue; }

	if (!strcmp(argv[count], "-data"))
		 { hp_format = DATA; continue; }

	if (!strcmp(argv[count], "-standard"))
		 { hp_system = STANDARD; continue; }

	if (!strcmp(argv[count], "-enhanced"))
		 { hp_system = ENHANCED; continue; }

	if (!strcmp(argv[count], "-color"))
		 { hp_system = COLOR; continue; }

	if (!strcmp(argv[count], "-info"))
		 { info = TRUE; continue; }

	if (!strncmp(argv[count], "-", 1) && (count < argc))
		 { print_usage(argv[0]); exit(0); }

	if (!infile) { infile = argv[count]; continue; }
	if (!outfile) { outfile = argv[count]; continue; }
  }
  
  if ((info && !infile) || (!info && !outfile)) { print_usage(argv[0]); exit(0); }
  	
  for (i=0; i<3; i++) {
  	ifp[i] = NULL;
  	ofp[i] = NULL;
  }
  
  /* open input file */
  if (!(ifp[0] = fopen(infile, "rb"))) {
	create_filename(tmp, infile, 1);
  	if (!(ifp[0] = fopen(tmp, "rb"))) {
		fprintf(stderr, "Could not open file \"%s\" or file \"%s\"\n", infile, tmp);
		exit(-1);
	}
  }

  fseek(ifp[0], 0, SEEK_SET);
  fread(buffer, 256, 1, ifp[0]);
  id = buffer[0]*256 + buffer[1];

  /* get system infos */
  switch (id) {
  case 0x424d:  	
	offset = *((DWORD *)(buffer + 10));
	break;
  case 0x0004:
	hp_format = BDAT;
	tablesize = buffer[10]*256 + buffer[11];
	break;
  case 0x000A:
	hp_format = DATA;
  	tablesize = 0;
    fseek(ifp[0], 0, SEEK_SET);
  	while (fread(&hpint, 1, 4, ifp[0]) == 4) {
  		if (hpint.id == 0x0a00) tablesize++;
  		else break;
  	}
  	break;
  default:
  	fprintf(stderr, "Unknown input file format\n");
  	exit(-1);
  }

  if (info) printf("\nInfo on file %s:\n\n", infile);

  switch (id) {
  case 0x424d:	/* BMP input */
  	
  				if (info) {
  					printf(" File type: Windows BMP\n");
  					printf(" Reported file size: %lu bytes\n", *((DWORD *)(buffer + 2)));
  					printf(" Dimensions: %ld by %ld pixel\n", *((LONG *)(buffer + 18)), *((LONG *)(buffer + 22)));
  					printf(" Color depth: %u bit(s) per pixel\n", *((WORD *)(buffer + 28)));
  					printf(" Compression: ");
  					switch (*((DWORD *)(buffer + 30))) {
  					case 0: printf("no compression\n"); break;
  					case 1: printf("run length encoded (4 bpp, unsupported)\n"); break;
  					case 2: printf("run length encoded (8 bpp, unsupported)\n"); break;
  					case 3: printf("bitfield (unsupported)\n"); break;
  					}
  					exit(0);
  				}
  		
				/* if color, open second & third output file */
  				if (hp_system == COLOR) {
					create_filename(tmp, outfile, 1);
					if (!(ofp[0] = fopen(tmp, "wb"))) {
						fprintf(stderr, "Could not open file \"%s\"\n", tmp);
						exit(-1);
					}
					create_filename(tmp, outfile, 2);
					if (!(ofp[1] = fopen(tmp, "wb"))) {
						fprintf(stderr, "Could not open file \"%s\"\n", tmp);
						exit(-1);
					}
					create_filename(tmp, outfile, 3);
					if (!(ofp[2] = fopen(tmp, "wb"))) {
						fprintf(stderr, "Could not open file \"%s\"\n", tmp);
						exit(-1);
					}
				} else {
  				/* open output file */
					if (strlen(outfile) > 6) outfile[6] = 0;
					if (!(ofp[0] = fopen(outfile, "wb"))) {
						fprintf(stderr, "Could not open file \"%s\"\n", outfile);
						exit(-1);
					}
				}
					
				
				/* start file conversion */
  				bmp2hp(hp_format, hp_system, ifp[0], ofp, offset);
  				break;
  				
  case 0x0004:  /* HP format input */
  case 0x000A:
				/* determine system type */
				switch(tablesize) {
				case 16381:
					hp_system = STANDARD;
					break;
				case 15925:
					hp_system = ENHANCED;
					break;
				//case 23835:
				//case 23940:
				case 15960:	// 152 rows
				case 15855: // 151 rows
					hp_system = COLOR;
					break;
				default:
					fprintf(stderr, "Warning: wrong table size on input file (%d)\n", tablesize);
				}
				
				if (info) {
					printf(" File type: ");
					switch(hp_system) {
					case STANDARD:
						printf("Standard graphics (9845B with 98750A monitor)\n");
						printf(" Color depth: 1 bit per pixel\n");
						break;
					case ENHANCED:
						printf("Enhancend monochrome graphics (9845B option 200 with 98780A monitor)\n");
						printf(" Color depth: 1 bit per pixel\n");
						break;
					case COLOR:
						printf("Color graphics (9845C with 98770A color monitor), 3 files total\n");
						printf(" Color depth: 3 bits per pixel\n");
						break;
					}
					if (hp_format == BDAT) printf(" Data format: BDAT (untyped integers)\n");
					if (hp_format == DATA) printf(" Data format: DATA (typed integers)\n");
					printf(" Integer table size (dimensioned): %d\n", tablesize);
					exit(0);
				}
				
				/* if color, open second & third input file */
  				if (hp_system == COLOR) {
					create_filename(tmp, infile, 2);
					if (!(ifp[1] = fopen(tmp, "rb"))) {
						fprintf(stderr, "Could not open file \"%s\"\n", tmp);
						exit(-1);
					}
					create_filename(tmp, infile, 3);
					if (!(ifp[2] = fopen(tmp, "rb"))) {
						fprintf(stderr, "Could not open file \"%s\"\n", tmp);
						exit(-1);
					}
				}
				
  				/* open output file */
				if (!(ofp[0] = fopen(outfile, "wb"))) {
					fprintf(stderr, "Could not open file \"%s\"\n", outfile);
					exit(-1);
				}

				/* start file conversion */
  				hp2bmp(hp_format, hp_system, ifp, ofp[0]);
  				break;
  				
  default:		fprintf(stderr, "Unknown input file format\n");
  }

  /* cleanup */
  for (i=0; i<3; i++) {
  	if (ifp[i]) fclose(ifp[i]);
  	if (ofp[i]) fclose(ofp[i]);
  }
  exit(0);
}
