/******************************************************************************

	ima2hpi.c

	Does the transformation from IMA disk images (esp. derived from TD0 images
	via	Teledisk) back to hpi disc images (raw binary image format).
	
	Use the -l option together with Sergey Erokhin's Teledisk TDCVT unpacker
	and Dietmar Harlos's FDI2IMA utility in order to reproduce clean binary
	images from Teledisk archives.
	
	Note: beta version, not thoroughly tested, no deep error processing.

	(c) 2006 by Ansgar Kueckes (kueckes@alice-dsl.net)

******************************************************************************/

#ifdef __WIN32__
#include <windows.h>
#endif

#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <stdlib.h>

#ifndef __WIN32__
#define BYTE unsigned char
#define WORD unsigned short
#define DWORD unsigned long
#endif

typedef struct {
	DWORD cylinder, head, record, recsize;
} Sector_ID;

void print_usage(char *command)
{
	printf("Usage: %s <input file> [<output file>]\n", command);
	printf("        [-c <cylinders>][-h <heads>][-s <sectors>]\n");
	printf("        [-r <rec_size>][-j <track junk>][-i <interleave>]\n");
	printf("        [-cs <cylinder skew>][-hs <head skew>][-sc <skewcount>]\n");
	printf("        [-cskip <cylinder skip>][-hskip <head skip>]\n");
	printf("        [-l <control file>][-m <matchlist>][-v]\n\n");
	printf("where\n");
	printf("        <input file>    IMA image file\n");
	printf("        <output file>   HP image file (default=\"image.bin\")\n");
	printf("        <cylinders>     # of valid cylinders (default=auto)\n");
	printf("        <heads>         # of valid heads resp. sides (default=auto)\n");
	printf("        <sectors>       # of valid sectors per track (default=auto)\n");
	printf("        <rec_size>      record resp. sector size of valid sectors (default=256)\n");
	printf("        <junk>          extra data at end of track (default=0)\n");
	printf("        <cylinder skew> cylinder-to-cylinder skew factor of original disk (default cs=0)\n");
	printf("        <head skew>     head-to-head skew factor of original disk (default hs=0)\n");
	printf("        <skewcount>     # tracks to be skewed (default=all)\n");
	printf("        <interleave>    n:1 interleave factor of original disk (default=auto)\n");
	printf("        <control file>  control transformation with TDCVT v0.11 output file\n");
	printf("        <matchlist>     list of known transformations of the form 'n which where which where ...'\n");
	printf("        -v              verbose mode\n");
	return;
}

DWORD get_dword(BYTE *byte)
{
	return (byte[0]<<24)+(byte[1]<<16)+(byte[2]<<8)+byte[3];
}

/*
	get block of size <rec_size> at byte position
*/
int get_record(FILE *fp, BYTE *buffer, DWORD position, int rec_size)
{
	int bytes_read;

	if (fseek(fp, position, SEEK_SET) < 0) {
	   fprintf(stderr, "Seek error #%d\n", errno);
	   return -1;
	}
	
	if ((bytes_read = fread((void *)buffer, 1, rec_size, fp)) != rec_size) {
	   if (ferror(fp)) {
	      fprintf(stderr, "Read error (%d bytes read)\n", bytes_read);
	      return -1;
	   }	   
	}
	return 0;
}

/*
	set up interleave tables
*/
void get_interleave_tables(int* abs2log, int* log2abs, int num_sectors, int interleave)
{
	int i, current;

	/* build absolute-to-logical table */
	for (i=0; i<num_sectors; i++) abs2log[i] = -1;
	current = 0;
	abs2log[current] = 0;
	for (i=1; i<num_sectors; i++) {
		current = (current + interleave) % num_sectors;
		while (abs2log[current]>=0) current = (current + 1) % num_sectors;
		abs2log[current] = i;
	}
	
	/* build logical-to-absolute table */
	if (log2abs) {
		for(i=0; i<num_sectors; i++)
			log2abs[abs2log[i]] = i;
	}
}

/* parser routines for control file */

BOOL parse_track(BYTE *line, DWORD *ret_cylinder, DWORD *ret_head, DWORD *ret_num_sectors)
{
	line[19]=0;
	line[32]=0;
	line[40]=0;
	*ret_cylinder = atoi(line+17);
	*ret_head = atoi(line+30);
	*ret_num_sectors = atoi(line+38);

	return TRUE;
}

BOOL parse_sector(BYTE *line, Sector_ID *id)
{
	line[15]=0;
	line[20]=0;
	line[26]=0;
	line[35]=0;
	id->cylinder = atoi(line+13);
	id->head = atoi(line+19);
	id->record = atoi(line+24);
	id->recsize = atoi(line+32);
	
	return TRUE;
}


int main(int argc, char *argv[])
{
	FILE *ifp, *ofp, *cfp;				/* file pointers */
	BYTE buffer[1024];					/* work buffer */
	int abs2log[256];					/* interleave table for transformation absolute->logical */
	int log2abs[256];					/* interleave table for transformation logical->absolute */
	DWORD offset;
	DWORD record = 0;
	DWORD location;
	DWORD from, to;
	DWORD num_tracks = 0;				/* total number of tracks */
	DWORD num_cylinders = 0;			/* total number of cylinders */
	DWORD num_heads = 0;				/* total number of heads/sides */
	DWORD num_sectors = 0;				/* sectors per track */
	DWORD dir_start = 0;				/* directory start position in records (LIF only) */
	DWORD dir_size = 0;					/* directory size in records (LIF only) */
	DWORD data_start = 0;				/* data start position in records (LIF only) */
	int i,j,count;
	int rec_size = 256;					/* record size (default 256) */
	char *outfile = "image.bin";		/* output file name */
	char *infile = 0;					/* input file name */
	char *control_file = 0;				/* control file name */
	int cskip = 0;						/* physical cylinder-to-cylinder skip in bytes */
	int hskip = 0;						/* physical head-to-head skip in bytes */
	int junk = 0;						/* sector junk at end of track in bytes */
	int cskew = -1;						/* cylinder-to-cylinder skew (-1 = autodetect) */
	int hskew = -1;						/* head-to-head skew (-1 = autodetect) */
	int skewcount = 0;					/* skew count limiter (0 = no limit) */
	int interleave = -1;				/* n:1 interleave factor (-1 = autodetect) */
	int max_records = 0;				/* total record count */
	int cyl, head, sect;				/* current cylinder, head, sector */
	int matchcount = 0;					/* number of entries in matchlist (0 = don't use matchlist) */
	BOOL verbose = 0;					/* produce detailed output */
	DWORD matchlist[2][10];				/* matchlist (max. 10 entries) */
	
	/* process command line */
	if (argc < 2) { print_usage(argv[0]); exit(0); }
	
	for ( count = 1; count < argc ; count++)
	{
		if (!strcmp(argv[count], "-c") && (count < (argc - 1)))
			{ count++; num_cylinders = strtoul(argv[count],0,0); continue; }
	
		if (!strcmp(argv[count], "-h") && (count < (argc - 1)))
			{ count++; num_heads = strtoul(argv[count],0,0); continue; }
	
		if (!strcmp(argv[count], "-s") && (count < (argc - 1)))
			{ count++; num_sectors = strtoul(argv[count],0,0); continue; }
	
		if (!strcmp(argv[count], "-r") && (count < (argc - 1)))
			{ count++; rec_size = strtoul(argv[count],0,0); continue; }
	
		if (!strcmp(argv[count], "-j") && (count < (argc - 1)))
			{ count++; junk = strtoul(argv[count],0,0); continue; }
	
		if (!strcmp(argv[count], "-cskip") && (count < (argc - 1)))
			{ count++; cskip = strtoul(argv[count],0,0); continue; }
	
		if (!strcmp(argv[count], "-hskip") && (count < (argc - 1)))
			{ count++; hskip = strtoul(argv[count],0,0); continue; }
	
		if (!strcmp(argv[count], "-cs") && (count < (argc - 1)))
			{ count++; cskew = strtoul(argv[count],0,0); continue; }
	
		if (!strcmp(argv[count], "-hs") && (count < (argc - 1)))
			{ count++; hskew = strtoul(argv[count],0,0); continue; }
	
		if (!strcmp(argv[count], "-sc") && (count < (argc - 1)))
			{ count++; skewcount = strtoul(argv[count],0,0); continue; }
	
		if (!strcmp(argv[count], "-i") && (count < (argc - 1)))
			{ count++; interleave = strtoul(argv[count],0,0); continue; }
	
		if (!strcmp(argv[count], "-l") && (count < (argc - 1)))
			{ count++; control_file = argv[count]; continue; }
	
		if (!strcmp(argv[count], "-m") && (count < (argc - 1))) {
			count++;
			matchcount = strtoul(argv[count],0,0);
			for (i=0; i<matchcount; i++) {
				count++; matchlist[0][i] = strtoul(argv[count],0,0);	/* record number */
				count++; matchlist[1][i] = strtoul(argv[count],0,0);	/* byte position */
			}
			continue;
		}
	
		if (!strcmp(argv[count], "-v") && (count < argc))
			{ verbose = 1; continue; }
	
		if (!strncmp(argv[count], "-", 1) && (count < argc))
			{ print_usage(argv[0]); exit(0); }
	
		if (!infile) infile = argv[count];
		else outfile = argv[count];
	}
	if (!infile) { print_usage(argv[0]); exit(0); }
	
	if (rec_size > 1024) { fprintf(stderr, "Sorry, max. rec_size is 1024 bytes.\n"); exit(0); }
	
	/* open files */
	ifp = fopen( infile, "rb");
	if (!ifp) { fprintf(stderr, "Could not open input file \"%s\".\n", infile); exit(-1); }
	
	ofp = fopen( outfile, "wb");
	if (!ofp) { fprintf(stderr, "Could not open output file \"%s\" for writing.\n", outfile); exit(-1); }
	
	if (control_file) {
		cfp = fopen( control_file, "rb");
		if (!cfp) { fprintf(stderr, "Could not open control file \"%s\".\n", control_file); exit(-1); }
	}
	
	/* get first record */
	if (get_record(ifp, buffer, 0, rec_size)<0) exit(-1);
		
	/* check for LIF format */
	if ((buffer[0]==0x80) && (buffer[1]==0x00)) {
		
	/* get directory info */
	dir_start = get_dword(buffer+8);
	dir_size = get_dword(buffer+16);
	data_start = dir_start + dir_size;
	
	/* get drive geometry */
	if (num_cylinders == 0) num_cylinders = get_dword(buffer+24);
	if (num_heads == 0) num_heads = get_dword(buffer+28);
	if (num_sectors == 0) num_sectors = get_dword(buffer+32);
		
	/* guess interleave factor */	
	if (interleave<0) {
		for (record = 1; record<data_start; record ++) {
			if (get_record(ifp, buffer, record*rec_size, rec_size)<0) exit(-1);
			location = get_dword(buffer+12);
			if (location == data_start) {
				interleave = record / dir_start;
				break;
			}
		}
		if (record == data_start) {
			printf("Warning: could not guess interleave. setting interleave=2.\n");
			interleave = 2;
		}
	}
	if (verbose) printf("LIF detected: dirstart=%lu dirsize=%lu datastart=%lu\n",
		dir_start, dir_size, data_start);
	}
	else {
		printf("Warning: no LIF file\n");
		printf("Unable to perform auto-detection of media parameters (using defaults)\n");

		/* set defaults*/
		if (num_cylinders == 0) num_cylinders = 33;
		if (num_heads == 0) num_heads = 2;
		if (num_sectors == 0) num_sectors = 16;
		if (interleave<0) interleave = 2;
	}
	if (verbose) printf("Using tracks=%lu heads=%lu sectors=%lu interleave=%d\n",
		num_cylinders, num_heads, num_sectors, interleave);
		
	/* control file drive processing  (works with TDCVT v0.11 output) */

	if (control_file) {
		DWORD phy_cylinder, phy_head;		/* physical cylinder & head count (as reported by control file) */
		BOOL new_track = FALSE;
		BYTE line[1024];					/* line buffer */
		BYTE track_buffer[256][1024];		/* track buffer (up to 256 sectors with 1024 bytes each) */
		Sector_ID id;						/* current sector data */
		int line_no = 1;					/* currently processed line number in control file */
		int start_sector = 256;				/* start_sector (derived from very first track) */
		int sector_size = 0;				/* valid sector size (derived from very first track) */
		int track_size = 0;					/* valid sectors per track (derived from very first track) */
		char *ret;
		
		printf("Using control file %s...", control_file);
		if (verbose) printf("\n");
		
		if (fseek(ifp, 0, SEEK_SET) < 0) {
		   fprintf(stderr, "Seek error #%d\n", errno);
		   return -1;
		}
		
		while(1) {	/* track loop */
			/* parse track info */
			if (!new_track) {
				ret = fgets(line, 1024, cfp);
				line_no++;
			}
			new_track = FALSE;
			if (ret == NULL) break;									/* EOF or read error */
			if (strncmp(line, "\xfe TRACK:", 8)) continue;			/* junk */
			if (parse_track(line, &phy_cylinder, &phy_head, &num_sectors)) {
				if ((track_size == 0) &&
					(phy_cylinder == 0) &&
					(phy_head == 0)) track_size = num_sectors;
				i = 0;
				while (1) { /* sector loop */
					/* parse sector info */
					ret = fgets(line, 1024, cfp);
					line_no++;
					if (ret == NULL) break;							/* EOF or read error */
					if (!strncmp(line, "\xfe TRACK:", 8)) {			/* next track */
						new_track = TRUE;
						break;
					}
					if (strncmp(line, " + Sector:", 10)) continue;	/* junk */
					if (parse_sector(line, &id)) {
		
						/* determine valid sector size */
						if ((sector_size == 0) &&
							(phy_cylinder == 0) &&
							(phy_head == 0) &&
							(i == 0)) sector_size = id.recsize;
		
						/* determine start sector (normally 0 or 1) */
						if ((phy_cylinder == 0) &&
							(phy_head == 0) &&
							(id.record < start_sector)) start_sector = id.record;
		
						/* skip junk data */
						if (num_sectors != track_size) {
							if (verbose) printf("num sectors=%lu doesn't match track size=%d\n",
								num_sectors, track_size);
							if (fread((void *)buffer, 1, id.recsize, ifp) != id.recsize) {
							   if (ferror(ifp)) fprintf(stderr, "Read error on input file\n");
							   else fprintf(stderr, "Premature end of input data (exiting)\n");
							   exit(-1);
							}
						}
						else if (id.cylinder != phy_cylinder) {
							if (verbose) printf("cylID=%lu doesn't match physical cylinder=%lu\n",
								id.cylinder, phy_cylinder);
							if (fread((void *)buffer, 1, id.recsize, ifp) != id.recsize) {
							   if (ferror(ifp)) fprintf(stderr, "Read error on input file\n");
							   else fprintf(stderr, "Premature end of input data (exiting)\n");
							   exit(-1);
							}
						}
						else if (id.head != phy_head) {
							if (verbose) printf("headID=%lu doesn't match physical head=%lu\n",
								id.head, phy_head);
							if (fread((void *)buffer, 1, id.recsize, ifp) != id.recsize) {
							   if (ferror(ifp)) fprintf(stderr, "Read error on input file\n");
							   else fprintf(stderr, "Premature end of input data (exiting)\n");
							   exit(-1);
							}
						}
						else if (id.recsize != sector_size) {
							if (verbose) printf("recsizeID=%lu doesn't match valid sector size=%d\n",
								id.recsize, sector_size);
							if (fread((void *)buffer, 1, id.recsize, ifp) != id.recsize) {
							   if (ferror(ifp)) fprintf(stderr, "Read error on input file\n");
							   else fprintf(stderr, "Premature end of input data (exiting)\n");
							   exit(-1);
							}
						}
						else if (id.record > num_sectors) {
							if (verbose) printf("record=%lu exceeds max record number=%lu",
								id.record, num_sectors);
							if (fread((void *)buffer, 1, id.recsize, ifp) != id.recsize) {
							   if (ferror(ifp)) fprintf(stderr, "Read error on input file\n");
							   else fprintf(stderr, "Premature end of input data (exiting)\n");
							   exit(-1);
							}
						}
						else {
							/* fill track buffer */
							if (fread((void *)track_buffer[id.record], 1, sector_size, ifp) != sector_size) {
							   if (ferror(ifp)) fprintf(stderr, "Read error on input file\n");
							   else fprintf(stderr, "Premature end of input data (exiting)\n");
							   exit(-1);
							}
							i++;
						}
					}
					else {
						fprintf(stderr, "Parse error in line %d\n", line_no);
						exit(-1);
					}
				} /* sector loop */
				/* write out track buffer */
				if (verbose) printf("Processing cylinder=%lu head=%lu num_sectors=%d\n", phy_cylinder, phy_head, i);
				for (j=0; j<i; j++)
					if (fwrite((void *)track_buffer[j + start_sector], 1, sector_size, ofp) != sector_size) {
					   	if (ferror(ofp)) {
						   	fprintf(stderr, "Write error on file \"%s\" #%d\n", outfile, errno);
					      	exit(-1);	
						}
					}
			}
			else {
				fprintf(stderr, "Parse error in line %d\n", line_no);
				exit(-1);
			}
		} /* track loop */
		printf("done.\n");
		if (ifp) fclose(ifp);
		if (ofp) fclose(ofp);
		if (cfp) fclose(cfp);
		exit(0);
	}
	
	/* parameter driven processing */

	max_records = num_cylinders * num_heads * num_sectors;
	if (skewcount == 0) skewcount = num_cylinders * num_heads;
		
	get_interleave_tables(abs2log, log2abs, num_sectors, interleave);
	
	/* use matchlist to calculate proper skew factors (brute force..) */	
	if (matchcount) {
		for (i=0; i<num_sectors; i++) {	/* cylinder skew */
			for (j=0; j<num_sectors; j++) { /* track skew */
				for (count=0; count<matchcount; count++) {
					from = matchlist[0][count];
				  	cyl = (from / (num_sectors * num_heads));
				  	head = (from / num_sectors) % num_heads;
				  	num_tracks = head + (cyl * num_heads);
				  	to = from % num_sectors;
	
			 		/* add interleave */
			 		to = log2abs[to];
	
			  		/* add skew */
		  			to = (to + (cyl * (i - ((num_heads - 1) * j))) + (((cyl * num_heads) + head) * j)) % num_sectors;
	
		  			/* add base */
			  		to = ((to + (num_tracks * num_sectors)) * rec_size) + (cyl * cskip) + (head * hskip) + (num_tracks * junk);
		  			if (matchlist[1][count] != to) break;
				}
				if (count == matchcount) {
					if (verbose) printf("found matching cskew=%d tskew=%d\n", i, j);
					if (cskew < 0) cskew = i;
					if (hskew < 0) hskew = j;
				}
			}
		}
	}
	
	if (cskew < 0) cskew = 0;
	if (hskew < 0) hskew = 0;
	
	if (verbose) printf("using cylinder skew=%d head skew=%d\n", cskew, hskew);
	
	for (record = 0; record < max_records; record++) {

		/* CHS transformation */
		cyl = (record / (num_sectors * num_heads));
		head = (record / num_sectors) % num_heads;
		sect = record % num_sectors;
		num_tracks = head + (cyl * num_heads);
		
		// 	printf("[%lu] cyl=%d head=%d sect=%d ->", record, cyl, head, sect);
	
		/* add interleave */
		sect = log2abs[sect];
		
		/* add skew */
		if ((cyl * num_heads) < skewcount)
			sect = (sect + (cyl * (cskew - ((num_heads - 1) * hskew))) + (((cyl * num_heads) + head) * hskew)) % num_sectors;
		
		/* calculate position on media */
		offset = ((sect + (num_tracks * num_sectors)) * rec_size) + (cyl * cskip) + (head * hskip) + (num_tracks * junk);

		// 	printf("[%lu] cyl=%d head=%d sect=%d\n", offset, cyl, head, sect);
	
		if (get_record(ifp, buffer, offset, rec_size)<0) exit(-1);
		   	
		if (fwrite((void *)buffer, 1, rec_size, ofp) != rec_size)
			{
		    	if (ferror(ofp)) {
		      	fprintf(stderr, "Write error on file \"%s\" #%d\n", outfile, errno);
		      	exit(-1);	
			}   
		}
	}
	
	if (ifp) fclose(ifp);
	if (ofp) fclose(ofp);
	
	return 0;
}
