/******************************************************************************

	rommerge.c

	Combines upper & lower ROM into one ROM block & lists ROM statistics like
	checksum etc.
	
	Takes inverted ROM images (as produced by a ROM dump) and outputs
	normal executable positive logic data.
	
	(c) 2006 by Ansgar Kueckes (kueckes@alice-dsl.net)

******************************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>

void print_usage(char *command)
{
	printf("Usage: %s <high ROM image> <low ROM image> <output file>\n", command);
	printf("       <high ROM image>  HP ROM image file (high order part)\n");
	printf("       <low ROM image>   HP ROM image file (low order part)\n");
	printf("       <output file>     file for combined image\n");
	return;
}


/* simulate ADA/ADB machine instruction */

void bpc_unsigned_add(unsigned long *reg, unsigned long summand, int *extend, int *overflow)
{
  long tmp,a,b;
  unsigned long result;

  /* check for overflow condition */  
  a=*reg; if (a>32767) a-=65536;
  b=summand; if (b>32767) b-=65536;
  tmp=a+b;
  if ((tmp<-32768) || (tmp>32767)) *overflow=1;

  /* integer add */
  result=(*reg)+summand;
  
  /* carry from bit 15 ? */
  if (result>65535) *extend=1;
  	
  *reg=result&0xFFFF;
}

void bpc_signed_add(long *reg, long summand, int *extend, int *overflow)
{
  long tmp,a,b,result;

  /* check for overflow condition */  
  tmp=(*reg)+summand;
  if ((tmp<-32768) || (tmp>32767)) *overflow=1;

  /* integer add */
  a=*reg; if (a<0) a+=65536; 
  b=summand; if (b<0) b+=65536;
  result=a+b;
  
  /* carry from bit 15 ? */
  if (result>65535) { result-=65536; *extend=1; }

  if (result>32767) result-=65536;
  *reg=result;
}

void bpc_rar(long *reg, int count)
{
  long a;
  int i,lsb;

  /* integer rotate right */
  a=*reg; if (a<0) a+=65536;
  for (i=0; i<count; i++) {
  	lsb=a&0x01;
  	a>>=1;
  	if (lsb) a|=0x8000;
  }

  if (a>32767) a-=65536;
  *reg=a;
}

void bpc_clear_bit15(long *reg)
{
  long a;
   
  a=*reg; if (a<0) a+=65536;
  a&=0x7FFF;
   
  //if (a>32767) a-=65536;
  *reg=a; 
}


int main(int argc, char *argv[])
{
  int count;
  char *outfile = 0;
  char *hi_file = 0;
  char *lo_file = 0;
  FILE *hfp, *lfp, *ofp;
  int blocksize=4096;
  unsigned char hi_buffer[4096];
  unsigned char lo_buffer[4096];
  unsigned char outbuffer[8192];
  size_t bytes_read;
  long checksum;
  long verify;
  int overflow, extend, excount, ovcount;
  long z;
  int i,j;

  if (argc < 2) { print_usage(argv[0]); exit(0); }

  /* command line processing */
  for ( count = 1; count < argc ; count++)
  {
	if (!strncmp(argv[count], "-", 1) && (count < argc))
		 { print_usage(argv[0]); exit(0); }

	if (!hi_file) { hi_file = argv[count]; continue; }
	if (!lo_file) { lo_file = argv[count]; continue; }
	if (!outfile) { outfile = argv[count]; continue; }
  }
  if ((!hi_file) || (!lo_file) || (!outfile)) { print_usage(argv[0]); exit(0); }
  
  /* open ROM image files */
  hfp = fopen( hi_file, "rb");
  if (!hfp) { fprintf(stderr, "Could not open file \"%s\".\n", hi_file); exit(-1); }
  
  lfp = fopen( lo_file, "rb");
  if (!lfp) { fprintf(stderr, "Could not open file \"%s\".\n", lo_file); exit(-1); }

  ofp = fopen( outfile, "wb");
  if (!ofp) { fprintf(stderr, "Could not open file \"%s\".\n", outfile); exit(-1); }


  for (i=0;;i++) {
        checksum=0;
        verify=0;
        overflow=0;
        excount=0;
        ovcount=0;
  	
	/* read in one single ROM block */
	if ((bytes_read = fread((void *)hi_buffer, 1, blocksize, hfp)) != blocksize) {
		if (ferror(hfp)) {
	   		fprintf(stderr, "Read error on file \"%s\" (%d bytes read)\n", hi_file, bytes_read);
			exit(-1);
		}
		exit(0);
	}
	if ((bytes_read = fread((void *)lo_buffer, 1, blocksize, lfp)) != blocksize) {
		if (ferror(lfp)) {
	   		fprintf(stderr, "Read error on file \"%s\" (%d bytes read)\n", lo_file, bytes_read);
			exit(-1);
		}	   
		exit(0);
	}
	/* merge ROMs */
	for (j=0; j<blocksize; j++) {
	  	outbuffer[j*2]=~hi_buffer[j];
	  	outbuffer[j*2+1]=~lo_buffer[j];
	  	z=outbuffer[j*2]*256+outbuffer[j*2+1];
	  	if (z>32767) z-=65536;
	  		
	  	/* update verify */ 
		extend=0;
                bpc_signed_add(&verify, z, &extend, &overflow);
        	while (extend) {
        		extend=0;
        		bpc_signed_add(&verify, 1, &extend, &overflow);
        		excount++;
        	}
        	
	  	/* update checksum */        	
        	bpc_rar(&checksum, 3);
        	if (checksum<0) {
        		//bpc_clear_bit15(&checksum);
        		bpc_rar(&checksum, 2);
        	}
		bpc_signed_add(&checksum, z, &extend, &overflow);
        	while (extend) {
        		extend=0;
        		bpc_signed_add(&checksum, 1, &extend, &overflow);
        		excount++;
        	}
	}
       	printf("block %d: verify=%ld checksum=%ld", i, verify, checksum);
	if (verify == -1) printf ("\t(ROM OK)\n"); else printf ("\t(ROM currupt)\n");
	
	if (fwrite((void *)outbuffer, 1, blocksize*2, ofp) != blocksize*2) {
		if (ferror(ofp)) {
			fprintf(stderr, "Write error on file \"%s\" #%d\n", outfile, errno);
			exit(-1);	
		}   
	}
  }
  
  /* cleanup */
  if (hfp) fclose(hfp);
  if (lfp) fclose(lfp);
  if (ofp) fclose(ofp);
  exit(0);
}
